/*  Funktionsweise
	- Hallo sagen
	- xmodem anstossen
	- intakte Pakete lesen
	  - Zeichen an Parser weiterleitena
		    - pages bilden  
			- diese brennen
	- Anwendung starten
Atmega328P: BootSZ1 und SZ0 = 0 - sprich geflasht ergibt 4096byte Bootloader
Einsprungadresse: -Ttext=0x7000
1kWord ergibt 0x7800

15Juli2015: HM_TRP-Enable nutzen

*/

//#include <stdio.h>	
//#include <inttypes.h>
#include <string.h>						// wegen memset
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/boot.h>

#define BAUD 	9600
#define E_PIN   PORTD6
#define HM_AN   PORTD&=~(1<<E_PIN)
#define HM_AUS  PORTD|=(1<<E_PIN)
                        
 /* Zustnde des Hex-File-Parsers */
#define PARSER_STATE_START      0
#define PARSER_STATE_SIZE       1
#define PARSER_STATE_ADDRESS    2
#define PARSER_STATE_TYPE       3
#define PARSER_STATE_DATA       4
#define PARSER_STATE_CHECKSUM   5
#define PARSER_STATE_ERROR      6
#define PARSER_STATE_EOF      	7
#define PARSER_STATE_END     	8


#define MAX_PROTO_ERRORS    10
#define XMODEM_PACKET_SIZE  128


//#define SPM_PAGESIZE       64   // 32 words atmega8, atmega88
//#define SPM_PAGESIZE      128   // 64 words atmega168
//#define SPM_PAGESIZE      256   // 128 words atmega128

#define ASC_TMO  0x00           /* Timeout special */
#define ASC_SOH  0x01           /* Start Of Header Ctrl-A */
#define ASC_EOT  0x04           /* End OF Transmission Ctrl-D */
#define ASC_ACK  0x06           /* Acknowledge Ctrl-F */
#define ASC_NAK  0x15           /* Not Acknowledge Ctrl-U */
#define ASC_CAN  0x18           /* Cancel Ctrl-X */
#define ASC_ERR  0xFF           /* Error special */
#define START_SIGN ':'      	/* Hex-Datei Zeilenstartzeichen */

// Timeout if zero.
volatile unsigned short intime;         // knnte uint16_t sein
volatile unsigned short bootime=15;
volatile unsigned short dval=0;

static unsigned char RecvPacket(unsigned char pn, unsigned char *cp);
static void SendOctet(char ch);
static unsigned char RecvOctet(void);
static void SendString(char *buffer);    // nur fr Prosa
static uint16_t hex2num (const uint8_t * ascii, uint8_t num);
static void Parser (uint8_t c);
void program_page (uint32_t page, uint8_t *buf);

static uint16_t hex_addr = 0,   /* Intel-HEX Zieladresse */
                flash_page = 0, /* Zu schreibende Flash-Page */  
                hex_check = 0,  /* Intel-HEX Checksumme zum berprfen des Daten */
                flash_cnt = 0;  /* Positions zum Schreiben in der Datenpuffer */
                    
static uint8_t  parser_state = PARSER_STATE_START,/* Empfangszustandssteuerung */
                flash_page_flag = 1,/* Flag zum ermitteln einer neuen Flash-Page */
                flash_data[SPM_PAGESIZE], /* Datenpuffer fr die Hexdaten*/
                hex_cnt = 0, /* Position zum Schreiben in den HEX-Puffer */
                hex_buffer[5], /* Puffer fr die Umwandlung der ASCII in Binrdaten */
                hex_size = 0,/* Intel-HEX Datenlnge */
                hex_data_cnt = 0, /* Zhler fr die empfangenen HEX-Daten einer Zeile */
                hex_type = 0, /* Intel-HEX Recordtype */
                hex_checksum=0; /* empfangene HEX-Checksumme */

int main(void)
{
    unsigned char ch;   /* Result of last packet receive or response. */
    unsigned char ec;   /* Error counter. */
    unsigned short pn;  /* Expected packet number. */
    unsigned char buff[XMODEM_PACKET_SIZE];  
    static uint8_t  temp,i;   /* temporre Variable */
    void     (*start)( void ) = 0x0000; 

    /* Fllen der Puffer mit definierten Werten */
    memset(hex_buffer, 0x00, sizeof(hex_buffer));
    memset(flash_data, 0xFF, sizeof(flash_data));
   /* Interrupt Vektoren verbiegen */
    temp = MCUCR;
    MCUCR = temp | (1<<IVCE);
    MCUCR = temp | (1<<IVSEL);

 	UBRR0L=(uint8_t)(F_CPU/(BAUD*16L))-1; 		// Baudrate festlegen
	UBRR0H=(uint8_t)((F_CPU/(BAUD*16L))-1)>>8; 	// Baudrate festlegen
	UCSR0B=(1<<RXEN0)|(1<<TXEN0);				// enable RX, TX 	
	DDRD=(1<<DDD5)|(1<<DDD6);					
	PORTD=(1<<PORTD5)|(1<<PORTD6);				// PD5-Config PD6-Enable PD7-Test fr HM-TRP
    sei();
    HM_AN; 
//	SendString("\r\nPlease start Xmodem transfer! \r\n");
	
	pn = 1; /* XMODEM starts with packet number one. */
    for (;;bootime--) {
        ch = RecvPacket(pn, buff);
        if (ch != ASC_TMO && ch != ASC_ERR) {
            break;
     		}
        if (bootime==0){			// nach BootTimeOut Anwendung starten
            SendString("TimeOut!\r\n");
 			break;	
			}
        _delay_ms(1000);
        SendOctet(ASC_NAK);
		}

    ec = 0;
    while (bootime>0) {  			// Process packets
        if (ch == ASC_SOH) {        // valid packet can be flashed
 	        for (i=0; i<XMODEM_PACKET_SIZE; i++){
		        Parser(buff[i]);		
                }
			if (parser_state==PARSER_STATE_ERROR){
				SendOctet(ASC_CAN);
 				SendString("\r\n Error occured!\r\n");
                break;               // Exit and start Applikation
 				}
			ec = 0; 				/* Clear error counter. */
            pn++;   				/* Increment packet number. */
            ch = ASC_ACK;  			/* Respond with ACK. -> nchstes Paket */
            }
        else if (ch == ASC_EOT || ch == ASC_CAN) {  // end-of-transmission or cancel.
 	        for (i=0; i<XMODEM_PACKET_SIZE; i++){
		        Parser(buff[i]);		
            	}
			if (parser_state==PARSER_STATE_ERROR){
				SendOctet(ASC_CAN);
 				SendString("\r\n Error occured!\r\n");
                break;               // Exit and start Applikation
 				}
            SendOctet(ASC_ACK);     // Respond with ACK. -> Schluss
            break;					// raus aus for -Schleife
            } 
        else {                      //  Anything else is treated as an error 
            if (ec++ > MAX_PROTO_ERRORS) {
                SendOctet(ASC_CAN);
                break;               // Exit and start Applikation
                }
            ch = ASC_NAK;            // Respond with NAK. -> Wiederholung 
            }                            
		
        SendOctet(ch);
        intime=0;
		while ( !(UCSR0A & (1<<RXC0)) ) {
            if (++intime == 0){        // warte auf Antwort?
                dval=1;
				break;
				}
            }                          
 		if (intime==0) 	SendOctet(ch); // Quittung verlorengegangen, also nochmal
        ch = RecvPacket(pn, buff);     // next or repeated packet 
	    }                              // Ende der While-Schleife
	SendString("Parser-State:");
    SendOctet(parser_state+48);
	if (dval==1) SendString("\r\nQuittung wurde wiederholt\r\n");
 	SendString("\r\nReset AVR!\r\n");
 
    /* Interrupt Vektoren wieder gerade biegen */
    temp = MCUCR;
    MCUCR = temp | (1<<IVCE);
    MCUCR = temp & ~(1<<IVSEL);
    /* Reset */
    start();
    return 0;
}

static uint16_t hex2num (const uint8_t * ascii, uint8_t num){ // 32bytes
    uint8_t  i;
    uint16_t val = 0;
    for (i=0; i<num; i++){    // Hex-Ziffer auf ihren Wert abbilden 
        uint8_t c = ascii[i];       
        if (c >= '0' && c <= '9')            c -= '0';  
        else if (c >= 'A' && c <= 'F')       c -= 'A' - 10;
        else if (c >= 'a' && c <= 'f')       c -= 'a' - 10;           
        val = 16 * val + c;
    }
    return val;  
}

static void SendString(char *buffer){
	for (int i=0; buffer[i] !=0;i++)
		SendOctet (buffer[i]);
}

static void SendOctet(char ch){
	while (!(UCSR0A & (1<<UDRE0)));	//warte bis UDR leer			
	UDR0=ch;							//senden
}

static unsigned char RecvOctet(void){ // 26bytes
    intime = 1;
    while ( !(UCSR0A & (1<<RXC0)) ) {
        if (++intime == 0)
            return ASC_TMO;
    }
    return UDR0;
}

/*
 * Receive a protocol packet.
 *
 * \param pn Expected packet number.
 * \param cp Pointer to packet buffer.
 *
 * \return - ASC_SOH if a packet has been received.
 *         - ASC_CAN if remote cancels transmission.
 *         - ASC_EOT if transmission finished.
 *         - ASC_ERR on packet format errors.
 *         - ASC_TMO on timeouts.
 *
 * \todo I think that this will fail if the remote missed a previously
 *       sent ACK.
 */

static unsigned char RecvPacket(unsigned char pn, unsigned char *cp)
{ // 672 bytes
    unsigned char rc; // Function result. 
    unsigned char ch; // Calculated checksum. 
    unsigned char i;

    for (;;) {										// Warte auf Spezielle Zeichen 
        if ((rc = RecvOctet()) == ASC_SOH) {		// SOH received -> weiter
            break;
            }
        if (!intime || rc == ASC_EOT || rc == ASC_CAN) { // TimeOut oder Cancel -> raus
            return rc;
           }
    }

    ch = RecvOctet();
    if (RecvOctet() + ch != 0xFF || ch != pn)
        rc = ASC_ERR;                             // Fehler -> trotzdem weiterlesen

    ch = 0;
    for (i = 0; i < 128; i++) {
        ch += (*cp++ = RecvOctet());
        if (!intime) {
            return ASC_TMO;
            }
        }

    if (ch != RecvOctet() || !intime)		// Checksumme lesen und prfen
        rc = ASC_ERR;

    return rc;
}

static void Parser (uint8_t c){ 			// 896bytes
       switch(parser_state){
           case PARSER_STATE_START:			// Warte auf Zeilen-Startzeichen 
                if((uint8_t)c == START_SIGN){
                    parser_state = PARSER_STATE_SIZE;
                    hex_cnt = 0;
                    hex_check = 0;
                    }
                break;
           case PARSER_STATE_SIZE:	// Parse Datengre 
                hex_buffer[hex_cnt++] = (uint8_t)c;
                if(hex_cnt == 2){
     			    parser_state = PARSER_STATE_ADDRESS;
	                hex_cnt = 0;
                    hex_size = (uint8_t)hex2num(hex_buffer, 2);
     				hex_check += hex_size;
                    }
                break;
           case PARSER_STATE_ADDRESS:  // Parse Zieladresse 
                hex_buffer[hex_cnt++] = (uint8_t)c;
                if(hex_cnt == 4){
                    parser_state = PARSER_STATE_TYPE;
                    hex_cnt = 0;
                    hex_addr = hex2num(hex_buffer, 4);
                    hex_check += (uint8_t) hex_addr;
                    hex_check += (uint8_t) (hex_addr >> 8);
                    if(flash_page_flag){
                       flash_page = hex_addr - hex_addr % SPM_PAGESIZE;
                       flash_page_flag = 0;
                       }
                    }
                break;
           case PARSER_STATE_TYPE:   // Parse Zeilentyp 	
                hex_buffer[hex_cnt++] = (uint8_t)c;
                if(hex_cnt == 2){
                   hex_cnt = 0;
                   hex_data_cnt = 0;
                   hex_type = (uint8_t)hex2num(hex_buffer, 2);
                   hex_check += hex_type;
                   switch(hex_type){
                       case 0: parser_state = PARSER_STATE_DATA; break;
                       case 1: parser_state = PARSER_STATE_CHECKSUM; break;
                       default: parser_state = PARSER_STATE_DATA; break;
                       }
                   }
                 break;
	        case PARSER_STATE_DATA:  // Parse Flash-Daten 
                 hex_buffer[hex_cnt++] = (uint8_t)c;
                 if(hex_cnt == 2){
                    hex_cnt = 0;
                    flash_data[flash_cnt] = (uint8_t)hex2num(hex_buffer, 2);
                    hex_check += flash_data[flash_cnt];
                    flash_cnt++;
                    hex_data_cnt++;
                    if(hex_data_cnt == hex_size){
                        parser_state = PARSER_STATE_CHECKSUM;
                        hex_data_cnt=0;
                        hex_cnt = 0;
                        }
                    if(flash_cnt == SPM_PAGESIZE){  // Puffer voll -> schreibe Page
                        program_page((uint16_t)flash_page, flash_data);
                        memset(flash_data, 0xFF, sizeof(flash_data));
                        flash_cnt = 0;
                        flash_page_flag = 1;
                        }
                     }
                  break;
             case PARSER_STATE_CHECKSUM:  // Parse Checksumme    
                  hex_buffer[hex_cnt++] = (uint8_t)c;
                  if(hex_cnt == 2){
                     hex_checksum = (uint8_t)hex2num(hex_buffer, 2);
                     hex_check += hex_checksum;
                     hex_check &= 0x00FF;
                     if(hex_type == 1){  // letzte Zeile  
                        if (flash_cnt==0)  // erste Page nicht berschreiben
					        parser_state = PARSER_STATE_END;
                     	else	
							parser_state = PARSER_STATE_EOF;
                     	}
                     else{         // berprfe Checksumme -> muss '0' sein 
                        if(hex_check == 0) parser_state = PARSER_STATE_START;
                        else parser_state = PARSER_STATE_ERROR;
                        }
					 }
                  break;			
             case PARSER_STATE_EOF:    //  schreibe Restdaten 
			      program_page((uint16_t)flash_page, flash_data);
  				  parser_state = PARSER_STATE_END;                    
                  break;			
	         case PARSER_STATE_END:    // Jetzt nur noch Fllzeichen vom Xmodem
			 	  break;			
			 case PARSER_STATE_ERROR:  // Parserfehler (falsche Checksumme) 
                  break;			
  			 default:
                  break;
             }
}

void program_page (uint32_t page, uint8_t *buf){ // 56 bytes
    uint16_t i;
    uint8_t sreg;
 
    sreg = SREG; cli();			// Disable interrupts 
    eeprom_busy_wait ();
    boot_page_erase (page);
    boot_spm_busy_wait ();      // Wait until the memory is erased. 
    for (i=0; i<SPM_PAGESIZE; i+=2){        
        uint16_t w = *buf++; w += (*buf++) << 8; // Set up little-endian word. 
        boot_page_fill (page + i, w);
    }
    boot_page_write (page);     // Store buffer in flash page.		
    boot_spm_busy_wait();       // Wait until the memory is written.
    boot_rww_enable ();// Reenable RWW-section again. We need this if we want 
  				      // to jump back to the application after bootloading
	SREG = sreg;// Re-enable interrupts (if they were ever enabled).
}
